from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.keras.backend import _preprocess_padding, _preprocess_conv2d_input, shape, image_data_format
from tensorflow.python.ops import array_ops

from dynastes.core import nn as dnn


def depthwise_conv2d_transpose(x,
                               kernel,
                               output_shape,
                               strides=(1, 1),
                               padding='valid',
                               data_format=None,
                               dilation_rate=(1, 1)):
    """2D depthwise deconvolution (i.e.
    transposed depthwise convolution).
    Arguments:
        x: Tensor or variable.
        kernel: kernel tensor.
        output_shape: 1D int tensor for the output shape.
        strides: strides tuple.
        padding: string, `"same"` or `"valid"`.
        data_format: string, `"channels_last"` or `"channels_first"`.
        dilation_rate: Tuple of 2 integers.
    Returns:
        A tensor, result of transposed 2D convolution.
    Raises:
        ValueError: if `data_format` is neither `channels_last` or
        `channels_first`.
    """

    if data_format is None:
        data_format = image_data_format()
    if data_format not in {'channels_first', 'channels_last'}:
        raise ValueError('Unknown data_format: ' + str(data_format))

    # `atrous_conv2d_transpose` only supports NHWC format, even on GPU.
    if data_format == 'channels_first' and dilation_rate != (1, 1):
        force_transpose = True
    else:
        force_transpose = False

    x, tf_data_format = _preprocess_conv2d_input(x, data_format, force_transpose)

    if data_format == 'channels_first' and tf_data_format == 'NHWC':
        output_shape = (output_shape[0], output_shape[2], output_shape[3],
                        output_shape[1])
    if output_shape[0] is None:
        output_shape = (shape(x)[0],) + tuple(output_shape[1:])

    if isinstance(output_shape, (tuple, list)):
        output_shape = array_ops.stack(list(output_shape))

    padding = _preprocess_padding(padding)
    if tf_data_format == 'NHWC':
        strides = (1,) + strides + (1,)
    else:
        strides = (1, 1) + strides

    x = dnn.depthwise_conv2d_transpose(x, kernel, output_shape, strides,
                                       padding=padding,
                                       data_format=tf_data_format)

    if data_format == 'channels_first' and tf_data_format == 'NHWC':
        x = array_ops.transpose(x, (0, 3, 1, 2))  # NHWC -> NCHW
    return x



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def large_compatible_negative(tensor_type):
    """Large negative number as Tensor.
    This function is necessary because the standard value for epsilon
    in this module (-1e9) cannot be represented using tf.float16
    Args:
      tensor_type: a dtype to determine the type.
    Returns:
      a large negative number.
    """
    if tensor_type == tf.float16:
        return tf.float16.min
    return -1e9



import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import custom_object_scope
from tensorflow.python.framework import test_util
from tensorflow.python.keras.testing_utils import layer_test

import dynastes as d
from dynastes.layers.random_layers import StatelessRandomNormalLike


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class StatelessRandomNormalLikeTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        with custom_object_scope(d.object_scope):
            layer_test(
                StatelessRandomNormalLike, input_shape=(5, 32, 3))

            layer = StatelessRandomNormalLike(channels=1)
            input = tf.random.normal(shape=[4,16,32,5])
            seed = tf.convert_to_tensor(np.random.randint(tf.int64.min, tf.int64.max, size=(4,4), dtype=np.int64))
            res = layer([input, seed])
            self.assertNotAllClose(input, res)
            seed = tf.convert_to_tensor([1,2,1,3])
            res = layer([input, seed])
            self.assertAllEqual(res[0], res[2])


import abc

import tensorflow as tf
import tensorflow.keras as tfk
import tensorflow.keras.layers as tfkl


def _call_masked(layer, inputs, training=None, mask=None, **kwargs):
    if layer.supports_masking:
        out = layer(inputs, training=training, mask=mask, **kwargs)
        out_mask = layer.compute_mask(inputs, mask, **kwargs)
    else:
        out = layer(inputs, training=training, **kwargs)
        out_mask = mask
    return out, out_mask


class GrowingGanModel(tfk.Model, abc.ABC):
    @abc.abstractmethod
    def interpolate_domain(self, x, y, interp):
        """
        Return result of interpolating between domains
        """

    @staticmethod
    def interpolate_mask(x_mask, y_mask, interp):

        @tf.function
        def interp_masks(_x_mask, _y_mask, _interp):

            if _interp <= 0.:
                return _x_mask
            elif _interp >= 1.:
                return _y_mask
            else:
                return tf.math.logical_and(_x_mask, _y_mask)

        if x_mask is None:
            return y_mask
        else:
            if y_mask is None:
                return x_mask
            else:
                return interp_masks(x_mask, y_mask, interp)


class GrowingGanGenerator(GrowingGanModel, abc.ABC):
    def __init__(self,
                 n_lods=0,
                 **kwargs):
        super(GrowingGanGenerator, self).__init__(**kwargs)
        self.n_lods = n_lods

    """
    @tf.function
    def get_images_ag(x, lod_in):

        def grow(x, res, lod_in):
            print("grow, res:", 2 ** res, ' lods left: ', lod_in)
            y = block(res, x)

            def get_lod_out():
                y_rgb = torgb(res, y)  # This stage's output in rgb
                if lod_in > lod_in:  # Currently lerping
                    x_rgb = upscale2d(torgb(res - 1, x))
                    z = lerp(y_rgb, x_rgb, lod_in - lod_in)
                else:  # No need to lerp at this stage
                    z = y_rgb
                return upscale2d(z, 2 ** lod_in)  ##Conform to output

            if lod_in > 0:
                if lod_in < lod_in:  # Lod decreases per rec until we're at lod_in
                    return grow(y, res + 1, lod_in - 1)
                else:  # If this stage should output
                    return get_lod_out()
            else:  # We're at lod_in 0 so we return the "lod_in-stack", this would be the innermost call of the recursion loop
                return get_lod_out()

        # Stylegan starts at lod_in 3 = 16x16
        print('resolution_log2', resolution_log2)
        images_out = grow(x, 3, resolution_log2 - 3)
        return images_out
    """

    @abc.abstractmethod
    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        """ Return processing block here """

    @abc.abstractmethod
    def get_to_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        """
        Return a layer that transforms output of
        gan_layer at this lod_in into a tensor that
        is compatible with your output domain
        The domain should be the same for every
        lod_in, but doesn't need to be RGB etc, if
        such a conversion is handled by the layer returned
        by get_conform_to_output_layer()
        Lerping between lods is handled in this "domain"
        """

    @abc.abstractmethod
    def get_upscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        """
        Return a layer that scales input from input_lod
        to output lod_in dimensions, this happens in
        the "domain" space

        caveats:
        Possibly you might have to perform cumprod
        on hparams strides, depending on your architecture
        if it's a simple one, you might just return Upsampling2D here

        """

    @abc.abstractmethod
    def get_conform_to_output_layer_by_lod(self, input_lod) -> tfkl.Layer:
        """
        Return a layer that scales/transforms input @ input_lod
        to conform exactly to targets
        """

    def _get_output(self, inputs, lod_in=None, training=None, **kwargs):

        def grow(gen: GrowingGanGenerator, x, current_lod):

            y_layer = gen.get_gan_layer_by_lod(current_lod)
            y = y_layer(x, training=training, **kwargs)

            lods_left = gen.n_lods - (current_lod + 1)

            def get_lod_output(x, y, current_lod):
                y_domain = gen.get_to_domain_layer_by_lod(current_lod)(y, training=training, **kwargs)
                if lod_in > lods_left:

                    x_domain = gen.get_to_domain_layer_by_lod(current_lod - 1)(x, training=training, **kwargs)

                    x_as_y_domain = gen.get_upscale_domain_layer_by_lod(current_lod - 1, current_lod)(x_domain,
                                                                                                      training=training,
                                                                                                      **kwargs)
                    z = gen.interpolate_domain(y_domain, x_as_y_domain, lod_in - lods_left)
                else:
                    z = y_domain

                return gen.get_conform_to_output_layer_by_lod(current_lod)(z, training=training, **kwargs)

            def ret_fn(x, y, lod_in, lods_left, current_lod):
                if lods_left > 0:
                    if lod_in < lods_left:
                        return grow(gen, y, current_lod=current_lod + 1)
                    else:
                        return get_lod_output(x, y, current_lod)
                else:
                    return get_lod_output(x, y, current_lod)

            return ret_fn(x, y, lod_in, lods_left, current_lod)

        return grow(self, x=inputs, current_lod=0)

    def _get_output_masked(self, inputs, lod_in=None, training=None, mask=None, **kwargs):

        def grow(gen: GrowingGanGenerator, x, current_lod, mask=None):

            y_layer = gen.get_gan_layer_by_lod(current_lod)
            y, y_mask = _call_masked(y_layer, x, training=training, mask=mask, **kwargs)

            lods_left = gen.n_lods - (current_lod + 1)

            def get_lod_output(x, y, y_mask=None, x_mask=None):
                y_domain_layer = gen.get_to_domain_layer_by_lod(current_lod)
                y_domain, y_domain_mask = _call_masked(y_domain_layer, y, training=training, mask=y_mask, **kwargs)

                if lod_in > lods_left:

                    x_domain_layer = gen.get_to_domain_layer_by_lod(current_lod - 1)
                    x_domain, x_domain_mask = _call_masked(x_domain_layer, x, training=training, mask=x_mask, **kwargs)

                    x_to_y_layer = gen.get_upscale_domain_layer_by_lod(current_lod - 1, current_lod)
                    x_as_y_domain, x_as_y_mask = _call_masked(x_to_y_layer, x_domain, training=training,
                                                              mask=x_domain_mask, **kwargs)

                    z = gen.interpolate_domain(y_domain, x_as_y_domain, lod_in - lods_left)
                    z_mask = gen.interpolate_mask(y_domain_mask, x_as_y_mask, lod_in - lods_left)
                else:
                    z = y_domain
                    z_mask = y_domain_mask

                r, r_mask = _call_masked(gen.get_conform_to_output_layer_by_lod(current_lod), z, training=training,
                                         mask=z_mask, **kwargs)
                return r

            if lods_left > 0:
                if lod_in < lods_left:
                    return grow(gen, y, current_lod=current_lod + 1, mask=y_mask)
                else:
                    return get_lod_output(x, y, y_mask=y_mask, x_mask=mask)
            else:
                return get_lod_output(x, y, y_mask=y_mask, x_mask=mask)

        return grow(self, x=inputs, current_lod=0, mask=mask)

    def call(self, inputs, lod_in=None, training=None, mask=None, **kwargs):
        """
        @param lod_in: value between 0. and self.n_lods-1
        @param kwargs: optional arguments passed to every layer on call
        """

        lod_in = tf.maximum(0., tf.minimum(self.n_lods - 1, (self.n_lods - 1) - tf.convert_to_tensor(lod_in)))

        if mask is None:
            return self._get_output(inputs=inputs,
                                    lod_in=lod_in,
                                    training=training,
                                    **kwargs)
        else:
            return self._get_output_masked(inputs=inputs,
                                           lod_in=lod_in,
                                           training=training,
                                           mask=mask, **kwargs)


class GrowingGanClassifier(GrowingGanModel, abc.ABC):
    def __init__(self,
                 n_lods=0,
                 **kwargs):
        super(GrowingGanClassifier, self).__init__(**kwargs)
        self.n_lods = n_lods

    @abc.abstractmethod
    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        """ Return processing block here """

    @abc.abstractmethod
    def get_from_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        """
        Return a layer that transforms input from
        domain into one that can consumed by gan_layer at
        this lod_in.
        The domain should be the same for every
        lod_in, but doesn't need to be RGB etc, if
        such a conversion is handled by the layer returned
        by get_conform_to_output_layer()
        Lerping between lods is handled in this "domain"
        """

    @abc.abstractmethod
    def get_downscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        """
        Return a layer that scales input from input_lod
        to output lod_in dimensions, this happens in
        the "domain" space

        caveats:
        Possibly you might have to perform cumprod
        on hparams strides, depending on your architecture
        if it's a simple one, you might just return AveragePooling2D here

        """

    @abc.abstractmethod
    def get_input_transform_layer_by_lod(self, lod) -> tfkl.Layer:
        """
        Return a layer that scales/transforms input
        to domain @ lod
        """

    def _get_predictions(self, inputs, lod_in, training=None, **kwargs):

        def grow(cls: GrowingGanClassifier, current_lod, lods_left):

            if lods_left > 0 and lod_in < lods_left:
                x = grow(cls, current_lod + 1, lods_left - 1)

                x = cls.get_gan_layer_by_lod(current_lod)(x,
                                                          training=training,
                                                          **kwargs)
            else:
                x = cls.get_input_transform_layer_by_lod(current_lod)(inputs,
                                                                      training=training,
                                                                      **kwargs)
                x = cls.get_from_domain_layer_by_lod(current_lod)(x,
                                                                  training=training,
                                                                  **kwargs)
                x = cls.get_gan_layer_by_lod(current_lod)(x,
                                                          training=training,
                                                          **kwargs)
            if lod_in > lods_left:
                next_x = cls.get_input_transform_layer_by_lod(current_lod - 1)(inputs,
                                                                               training=training,
                                                                               **kwargs)
                next_x = cls.get_from_domain_layer_by_lod(current_lod - 1)(next_x,
                                                                           training=training,
                                                                           **kwargs)
                return cls.interpolate_domain(x, next_x, lod_in - lods_left)
            else:
                return x

        return grow(self, 1, self.n_lods - 1)

    def call(self, inputs, lod_in=None, training=None, mask=None, **kwargs):
        """
        @param lod_in: value between 0. and self.n_lods-1
        @param kwargs: optional arguments passed to every layer on call
        """

        lod_in = tf.maximum(0., tf.minimum(self.n_lods - 1, (self.n_lods - 1) - tf.convert_to_tensor(lod_in)))

        if mask is None:
            return self._get_predictions(inputs=inputs,
                                         lod_in=lod_in,
                                         training=training,
                                         **kwargs)


class SimpleGrowingGanClassifer(GrowingGanClassifier, abc.ABC):

    @abc.abstractmethod
    def get_gan_layer(self, layer_idx) -> tfkl.Layer:
        """
        Return your GAN processing layer
        Latent space needs to be strictly increasing or equal
        @param layer_idx: layer id 0...( n_lods - 1 )
        @type layer_idx: int
        """

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        if lod < 0:
            raise ArithmeticError
        return self.get_gan_layer((self.n_lods - lod))

    @abc.abstractmethod
    def get_from_domain_layer(self, layer_idx) -> tfkl.Layer:
        """
        Return a layer that converts from your domain-space
        to the latent space at 'layer_idx' index
        Latent space needs to be strictly increasing or equal
        NOTE! This requires one more than your GAN-layers
        @param layer_idx: layer id 0...(n_lods)
        @type layer_idx: int
        """

    def get_from_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.get_from_domain_layer((self.n_lods - lod))

    @abc.abstractmethod
    def get_downscale_domain_by_range(self, input_idx, output_idx) -> tfkl.Layer:

        """
        Return a "in-domain" downscaling layer between your scales, such as:
        self.strides = [2,3,3]
        strides = self.strides[input_idx : output_idx]
        if len(strides) > 0:
            scale = np.cumprod(strides)[-1]
            if scale > 1:
                return tfkl.AveragePooling2D((scale, scale), strides=(scale, scale), padding='same')
        return tfkl.Activation('linear')

        @param input_idx:
        @type input_idx: int
        @param output_idx:
        @type output_idx: int
        """

    def get_downscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        if input_lod - output_lod == 1:
            return tfkl.Activation('linear')
        return self.get_downscale_domain_by_range(self.n_lods - input_lod, self.n_lods - output_lod)

    def get_input_transform_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.get_downscale_domain_layer_by_lod(self.n_lods, lod)


class SimpleGrowingGanGenerator(GrowingGanGenerator, abc.ABC):

    @abc.abstractmethod
    def get_gan_layer(self, layer_idx) -> tfkl.Layer:
        """
        Return your GAN processing layer
        Latent space needs to be strictly decreasing or equal
        @param layer_idx: layer id 0...( n_lods - 1 )
        @type layer_idx: int
        """

    def get_gan_layer_by_lod(self, lod) -> tfkl.Layer:
        if lod > (self.n_lods - 1):
            return tfkl.Activation('linear')
        return self.get_gan_layer((self.n_lods - 1) - lod)

    @abc.abstractmethod
    def get_to_domain_layer(self, layer_idx) -> tfkl.Layer:
        """
        Return a layer that converts from your latent space
        to your domain space (RGB, what-have-you)
        NOTE! This requires one more than your GAN-layers
        @param layer_idx: layer id 0...(n_lods)
        @type layer_idx: int
        """

    def get_to_domain_layer_by_lod(self, lod) -> tfkl.Layer:
        return self.get_to_domain_layer((self.n_lods - 1) - lod)

    @abc.abstractmethod
    def get_upscale_domain_by_range(self, input_idx, output_idx) -> tfkl.Layer:

        """
        Return a "in-domain" upscale layer between your scales

        Example:

        self.strides = [2,3,3]
        strides = self.strides[input_idx : output_idx]
        if len(strides) > 0:
            scale = np.cumprod(strides)[-1]
            if scale > 1:
                return tfkl.UpSampling2D(size=(scale, scale))
        return tfkl.Activation('linear')

        @param input_idx: starting layer requested 0 ... (n_lods-1)
        @type input_idx: int
        @param output_idx: ending layer requested 0 ... (n_lods-1)
        @type output_idx: int
        """

    def get_upscale_domain_layer_by_lod(self, input_lod, output_lod) -> tfkl.Layer:
        if output_lod == input_lod:
            return tfkl.Activation('linear')
        input_idx = max((self.n_lods - 1) - output_lod, 0)
        output_idx = max((self.n_lods - 1) - input_lod, 0)
        return self.get_upscale_domain_by_range(input_idx, output_idx)

    def get_conform_to_output_layer_by_lod(self, input_lod) -> tfkl.Layer:
        return self.get_upscale_domain_layer_by_lod(input_lod, self.n_lods)



import tensorflow as tf
import tensorflow.keras.layers as tfkl
from dynastes.ops.t2t_common import shape_list

class StatelessRandomNormalLike(tfkl.Layer):

    """
    A layer normal producing layer that outputs shape like input
    with optional seed of lower frequency, ie:
    inputs[0] = tf.Tensor of shape [b, x, y, ch]
    inputs[1] = tf.Tensor (int32/int64) of shape [b, x//4]

    Useful for cases where you need deterministic random
    such as inverse-autoregressive flows or StyleGAN noise
    """

    def __init__(self,
                 mean=0.0,
                 stddev=1.0,
                 channels=-1,
                 **kwargs):
        super(StatelessRandomNormalLike, self).__init__(**kwargs)
        self.mean = mean
        self.stddev = stddev
        self.channels = channels

    def build(self, input_shape):
        super(StatelessRandomNormalLike, self).build(input_shape)

    @tf.function
    def _compute_shape(self, inputs, seed):
        inputs_shape = shape_list(inputs)
        if self.channels > 0:
            inputs_shape[-1] = self.channels
        seed_shape = shape_list(seed)
        seed_shape = seed_shape + ([1] * (len(inputs_shape) - len(seed_shape)))
        random_shape = [a // b for a, b in zip(inputs_shape, seed_shape)]
        while random_shape[0] == 1:
            random_shape = random_shape[1:]
        return tf.stack(random_shape)

    def _recurse_generate(self, seed, shape):
        seed_shape = shape_list(seed)
        if len(seed_shape) == 0:
            if seed.dtype == tf.int64:
                seed = tf.bitcast(seed, tf.int32)
            elif seed.dtype == tf.int32:
                seed = tf.bitcast(seed, tf.int16)
                seed = tf.cast(seed, tf.int32)
            return tf.random.stateless_normal(mean=self.mean, stddev=self.stddev, seed=seed, shape=shape)
        else:
            return tf.stack([self._recurse_generate(s, shape) for s in tf.unstack(seed)])

    def _stateless_random_normal(self, inputs, seed=None):
        inputs_shape = shape_list(inputs)
        if self.channels > 0:
            inputs_shape[-1] = self.channels
        if seed is None:
            return tf.random.normal(mean=self.mean, stddev=self.stddev, shape=inputs_shape)
        else:
            random_shape = self._compute_shape(inputs, seed)
            out = self._recurse_generate(seed, random_shape)
            return tf.reshape(out, inputs_shape)

    def call(self, inputs, training=None, mask=None):
        if type(inputs) == list:
            seed = inputs[1]
            inputs = inputs[0]
        else:
            seed = None
        return self._stateless_random_normal(inputs, seed)

    def compute_output_shape(self, input_shape):
        if type(input_shape) == list:
            inputs_shape = input_shape[0]
        else:
            inputs_shape = input_shape
        if self.channels > 0:
            inputs_shape[-1] = self.channels
        return inputs_shape

    def get_config(self):
        config = {
            'mean': self.mean,
            'stddev': self.stddev,
            'channels': self.channels,
        }
        base_config = super(StatelessRandomNormalLike, self).get_config()
        return {**base_config, **config}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from functools import partial

from tensorflow.python.keras import activations

from dynastes.blocks import layer_factory
from dynastes.layers import ActivatedKernelBiasBaseLayer


class SelfAttentionBlock1D(ActivatedKernelBiasBaseLayer):

    def __init__(self,
                 attention_dim,
                 output_dim,
                 kernel_size,
                 attention_type='LocalizedAttentionLayer1D',
                 q_type='Conv1D',
                 k_type=None,
                 v_type=None,
                 num_heads=1,
                 multiquery_attention=False,
                 depth_multiplier=1,
                 strides=1,
                 dilation_rate=1,
                 grouped=False,
                 group_size=1,
                 padding='same',
                 activation=None,
                 use_bias=False,
                 relative=False,
                 local=False,
                 sparse=False,
                 masked=False,
                 dropout_rate=0.,
                 max_relative_position=None,
                 lsh_bucket_length=4,
                 block_length=128,
                 filter_width=100,
                 mask_right=False,
                 add_relative_to_values=False,
                 **kwargs):
        super(SelfAttentionBlock1D, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)
        self.q_type = q_type

        if k_type is None:
            self.k_type = q_type
        else:
            self.k_type = k_type

        if v_type is None:
            self.v_type = q_type
        else:
            self.v_type = v_type

        self.attention_dim = attention_dim
        self.output_dim = output_dim
        self.kernel_size = kernel_size
        self.num_heads = num_heads
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.depth_multiplier = depth_multiplier
        self.padding = padding
        self.grouped = grouped
        self.group_size = group_size
        self.multiquery_attention = multiquery_attention
        self.relative = relative
        self.local = local
        self.masked = masked
        self.sparse = sparse
        self.dropout_rate = dropout_rate
        self.max_relative_position = max_relative_position
        self.lsh_bucket_length = lsh_bucket_length
        self.block_length = block_length
        self.filter_width = filter_width
        self.mask_right = mask_right
        self.add_relative_to_values = add_relative_to_values

        conv_partial = partial(layer_factory.get_1d_layer(kernel_size=kernel_size,
                                                          grouped=grouped,
                                                          group_size=group_size,
                                                          depth_multiplier=depth_multiplier,
                                                          padding=padding,
                                                          activation=None,
                                                          use_bias=True,
                                                          kernel_initializer=self.get_initializer('kernel'),
                                                          bias_initializer=self.get_initializer('bias'),
                                                          kernel_regularizer=self.get_regularizer('kernel'),
                                                          bias_regularizer=self.get_regularizer('bias'),
                                                          activity_regularizer=None,
                                                          kernel_constraint=self.get_constraint('kernel'),
                                                          bias_constraint=self.get_constraint('bias')))
        q_filters = attention_dim
        k_filters = attention_dim
        v_filters = output_dim

        if multiquery_attention:
            k_filters /= num_heads
            v_filters /= num_heads

        self.q_layer = conv_partial(type=q_type,
                                    filters=q_filters,
                                    stride=strides,
                                    dilation=dilation_rate, name='Conv-Q')
        self.k_layer = conv_partial(type=k_type,
                                    filters=k_filters,
                                    stride=1,
                                    dilation=1, name='Conv-K')
        self.v_layer = conv_partial(type=v_type,
                                    filters=v_filters,
                                    stride=1,
                                    dilation=1, name='Conv-V')

        attention_padding = padding
        self.attention_layer = layer_factory.get_1D_attention_layer(
            type=attention_type,
            strides=strides,
            dilation_rate=dilation_rate,
            num_heads=num_heads,
            padding=attention_padding,
            multiquery_attention=multiquery_attention,
            preshaped_q=True,
            relative=relative,
            local=local,
            masked=masked,
            sparse=sparse,
            dropout_rate=dropout_rate,
            max_relative_position=max_relative_position,
            lsh_bucket_length=lsh_bucket_length,
            block_length=block_length,
            mask_right=mask_right,
            filter_width=filter_width,
            add_relative_to_values=add_relative_to_values,
        )

    def build(self, input_shape):
        self.build_bias(self.output_dim)

    def call(self, inputs, training=None, mask=None, **kwargs):
        x = inputs
        q = self.q_layer(x, training=training)
        k = self.k_layer(x, training=training)
        v = self.v_layer(x, training=training)

        if mask is not None:
            mask = [mask, mask, mask]
        x, weights = self.attention_layer([q, k, v], mask=mask, training=training)

        return super(SelfAttentionBlock1D, self).call(x, **kwargs), weights

    def get_config(self):
        config = {
            'q_type': self.q_type,
            'k_type': self.k_type,
            'v_type': self.v_type,
            'multiquery_attention': self.multiquery_attention,
            'attention_dim': self.attention_dim,
            'output_dim': self.output_dim,
            'depth_multiplier': self.depth_multiplier,
            'kernel_size': self.kernel_size,
            'num_heads': self.num_heads,
            'strides': self.strides,
            'dilation_rate': self.dilation_rate,
            'padding': self.padding,
            'grouped': self.grouped,
            'group_size': self.group_size,
            'relative': self.relative,
            'local': self.local,
            'masked': self.masked,
            'sparse': self.sparse,
            'dropout_rate': self.dropout_rate,
            'max_relative_position': self.max_relative_position,
            'lsh_bucket_length': self.lsh_bucket_length,
            'block_length': self.block_length,
            'filter_width': self.filter_width,
            'mask_right': self.mask_right,
            'add_relative_to_values': self.add_relative_to_values,
        }
        base_config = super(SelfAttentionBlock1D, self).get_config()
        return {**base_config, **config}



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from tensorflow.python.keras.utils.generic_utils import deserialize_keras_object
from tensorflow.python.keras.utils.generic_utils import serialize_keras_object

from .spectral import SpectralNormalization

object_scope = {
    'SpectralNormalization': SpectralNormalization,
}


def get(identifier):
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        return deserialize(identifier)
    elif isinstance(identifier, six.string_types):
        identifier = str(identifier)
        # We have to special-case functions that return classes.
        # TODO(omalleyt): Turn these into classes or class aliases.
        if identifier == 'spectral':
            return SpectralNormalization()
    elif callable(identifier):
        return identifier
    else:
        raise ValueError('Could not interpret regularizer identifier:', identifier)


def serialize(regularizer):
    return serialize_keras_object(regularizer)


def deserialize(config, custom_objects={}):
    custom_objects = {**custom_objects, **object_scope}
    return deserialize_keras_object(
        config,
        module_objects=globals(),
        custom_objects=custom_objects,
        printable_module_name='regularizer')


# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.keras.regularizers import Regularizer
from tensorflow.keras.utils import custom_object_scope
from tensorflow.python.keras import regularizers as tf_regularizers
from tensorflow.python.keras.regularizers import deserialize as _deserialize
from tensorflow.python.keras.regularizers import serialize as _serialize

from . import orthogonal
from .orthogonal import Orthogonal

object_scope = {
    'Orthogonal': Orthogonal
}


class ModifyingRegularizer(Regularizer):

    def __init__(self, first, second,
                 fn,
                 **kwargs):
        super(ModifyingRegularizer).__init__(**kwargs)
        self.first = get(first)
        self.second = get(second)
        self.fn = fn

    def __call__(self, x):
        return self.fn(self.first(x), self.second(x))

    def get_config(self):
        return {'first': serialize(self.first),
                'second': serialize(self.second)}


def _add(self, other):
    return ModifyingRegularizer(self, other, lambda x, y: x + y)


def _sub(self, other):
    return ModifyingRegularizer(self, other, lambda x, y: x - y)


setattr(Regularizer, '__add__', _add)
setattr(Regularizer, '__sub__', _sub)


def serialize(regularizer):
    return _serialize(regularizer)


def deserialize(config, custom_objects={}):
    custom_objects = {**custom_objects, **object_scope}
    return _deserialize(config, custom_objects)


def get(regularizer):
    if type(regularizer) == str:
        if regularizer == 'orthogonal':
            return Orthogonal()
    with custom_object_scope(object_scope):
        return tf_regularizers.get(regularizer)


# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



import timeit

import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import custom_object_scope
from tensorflow.python.framework import test_util

import dynastes as d
from dynastes.layers.t2t_attention_layers import Attention1D, Attention2D


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))
        testCase.assertAllInRange(grad, -400., 400)


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class T2TAttention1DTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        with custom_object_scope(d.object_scope):
            t_steps = 12
            bs = 2
            dim = 16
            dim_mq = 32
            num_heads = 4
            s = 2

            layers = [
                (
                    'Local Masked',
                    Attention1D(num_heads=num_heads, self_attention=True, local=True, masked=True, block_length=8,
                                filter_width=6),
                    {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
                (
                    'Local Multiquery Masked',
                    Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                                masked=True, block_length=8,
                                filter_width=6),
                    {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                     'dim_v': dim_mq // num_heads}),
                (
                    'Sparse Unmasked',
                    Attention1D(num_heads=num_heads, self_attention=True, sparse=True, lsh_bucket_length=3,
                                mask_right=True),
                    {'self': True, 'steps_q': 32, 'steps_kv': 32, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
                (
                    'Normal',
                    Attention1D(num_heads=num_heads, self_attention=False),
                    {'self': False, 'steps_q': 16, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
                ),
                (
                    'Normal Multiquery',
                    Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=False),
                    {'self': False, 'steps_q': 16, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                     'dim_v': dim_mq // num_heads}
                ),
                (
                    'Relative',
                    Attention1D(num_heads=num_heads, self_attention=True, relative=True, max_relative_position=8,
                                mask_right=True),
                    {'self': True, 'steps_q': 16, 'steps_kv': 16, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
                ),
                (
                    'Relative Multiquery',
                    Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, relative=True,
                                max_relative_position=8,
                                mask_right=True),
                    {'self': True, 'steps_q': 32, 'steps_kv': 32, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                     'dim_v': dim_mq // num_heads}
                ),
                (
                    'Relative Masked',
                    Attention1D(num_heads=num_heads, self_attention=True, relative=True, max_relative_position=8,
                                masked=True,
                                mask_right=True),
                    {'self': True, 'steps_q': 16, 'steps_kv': 16, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
                ),
                (
                    'Relative Multiquery Masked',
                    Attention1D(num_heads=num_heads, multiquery_attention=True,
                                self_attention=True, relative=True, max_relative_position=8,
                                masked=True,
                                mask_right=True),
                    {'self': True, 'steps_q': 16, 'steps_kv': 16, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                     'dim_v': dim_mq // num_heads}
                ),
                (
                    'Local',
                    Attention1D(num_heads=num_heads, self_attention=True, local=True, block_length=8, filter_width=6),
                    {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
                (
                    'Local Multiquery',
                    Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                                block_length=8, filter_width=6),
                    {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                     'dim_v': dim_mq // num_heads}),


            ]

            def test_layer(layer, params):
                q = to_tensor(normal(size=(bs, params['steps_q'], params['dim_q']))
                              .astype(np.float32))
                k = to_tensor(normal(size=(bs, params['steps_kv'], params['dim_k']))
                              .astype(np.float32))
                v = to_tensor(normal(size=(bs, params['steps_kv'], params['dim_v']))
                              .astype(np.float32))
                mask_len = params['steps_q'] // 6
                mask_q = to_tensor(([True] * (params['steps_q'] - mask_len)) + ([False] * (mask_len)))
                mask_q = tf.expand_dims(mask_q, axis=0)
                mask_q = tf.tile(mask_q, [bs, 1])

                mask_kv = to_tensor(([True] * (params['steps_kv'] - mask_len)) + ([False] * (mask_len)))
                mask_kv = tf.expand_dims(mask_kv, axis=0)
                mask_kv = tf.tile(mask_kv, [bs, 1])

                def get_test_fn(layer, mask):
                    @tf.function
                    def test_func(q, k, v):
                        r, _ = layer([q, k, v], mask=mask)
                        return r

                    return test_func

                mask = [mask_q, mask_kv, mask_kv]
                r, _ = layer([q, k, v], mask=mask)
                test_fn = get_test_fn(layer, mask)
                # Warmup
                rfn = test_fn(q, k=k, v=v)
                self.assertAllClose(r, rfn)
                self.assertAllInRange(r, -200., 200)

                def fn():
                    test_fn(q, k=k, v=v)

                time = timeit.timeit(fn, number=2) / (params['steps_q'] * params['steps_kv'] * params['dim_q'])
                time *= 8192

                _test_grads(self, test_fn, [q, k, v])
                return time

            for (type, layer, params) in layers:
                print(type, test_layer(layer, params))


class T2TAttention2DTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        with custom_object_scope(d.object_scope):
            t_steps = 12
            bs = 2
            dim = 8
            dim_mq = 16
            num_heads = 4
            s = 2

            layers = [

                (
                    'Relative',
                    Attention2D(num_heads=num_heads, self_attention=True, relative=True, max_relative_position=8,
                                mask_right=True),
                    {'self': True, 'steps_q': 8, 'steps_kv': 8, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
                ),
                (
                    'Relative Multiquery',
                    Attention2D(num_heads=num_heads, multiquery_attention=True, self_attention=True, relative=True,
                                max_relative_position=8,
                                mask_right=True),
                    {'self': True, 'steps_q': 8, 'steps_kv': 8, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                     'dim_v': dim_mq // num_heads}
                )
            ]

            #TODO: Add these back but for 2D
            """
            (
                'Local',
                Attention1D(num_heads=num_heads, self_attention=True, local=True, block_length=8, filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
            (
                'Local Multiquery',
                Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                            block_length=8, filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}),
            (
                'Local Masked',
                Attention1D(num_heads=num_heads, self_attention=True, local=True, masked=True, block_length=8,
                            filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
            (
                'Local Multiquery Masked',
                Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                            masked=True, block_length=8,
                            filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}),
             """

        def test_layer(layer, params):
            q = to_tensor(normal(size=(bs, params['steps_q'], params['steps_q'], params['dim_q']))
                          .astype(np.float32))
            k = to_tensor(normal(size=(bs, params['steps_kv'], params['steps_kv'], params['dim_k']))
                          .astype(np.float32))
            v = to_tensor(normal(size=(bs, params['steps_kv'], params['steps_kv'], params['dim_v']))
                          .astype(np.float32))

            # mask_q = to_tensor(([True] * (params['steps_q'] - 3)) + ([False] * (3)))
            # mask_q = tf.expand_dims(mask_q, axis=0)
            # mask_q = tf.tile(mask_q, [bs, 1])

            # mask_kv = to_tensor(([True] * (params['steps_kv'] - 3)) + ([False] * (3)))
            # mask_kv = tf.expand_dims(mask_kv, axis=0)
            # mask_kv = tf.tile(mask_kv, [bs, 1])

            def get_test_fn(layer, mask):
                @tf.function
                def test_func(q, k, v):
                    r, _ = layer([q, k, v], mask=mask)
                    return r

                return test_func

            # mask = [mask_q, mask_kv, mask_kv]
            mask = None
            r, _ = layer([q, k, v], mask=mask)
            test_fn = get_test_fn(layer, mask)
            # Warmup
            rfn = test_fn(q, k=k, v=v)
            self.assertAllClose(r, rfn)
            self.assertAllInRange(r, -200., 200)

            def fn():
                test_fn(q, k=k, v=v)

            time = timeit.timeit(fn, number=2) / (
                        params['steps_q'] * params['steps_q'] * params['steps_kv'] * params['steps_kv'] * params[
                    'dim_q'])
            time *= 8192

            _test_grads(self, test_fn, [q, k, v])
            return time

        for (type, layer, params) in layers:
            print(type, test_layer(layer, params))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from dynastes import blocks
from dynastes import core
from dynastes import helpers
from dynastes import layers
from dynastes import models
from dynastes import ops
from dynastes import regularizers
from dynastes import util
from dynastes import weight_normalizers
from dynastes.core import backend
from dynastes import probability
from dynastes.probability import bijectors, bijector_partials

object_scope = {
    **blocks.object_scope,
    **layers.object_scope,
    **weight_normalizers.object_scope,
    **regularizers.object_scope,
}

__version__ = "0.2.8"
# Cleanup symbols to avoid polluting namespace.
del absolute_import
del division
del print_function



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow.python.keras import activations

from dynastes.layers import ActivatedKernelBiasBaseLayer
from dynastes.ops.time_delay_ops import time_delay_nn_1d


class _TimeDelayLayer(ActivatedKernelBiasBaseLayer):
    def __init__(self,
                 activation=None,
                 use_bias=True,
                 **kwargs):
        super(_TimeDelayLayer, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)

    def get_config(self):
        config = {}
        base_config = super(_TimeDelayLayer, self).get_config()
        return {**base_config, **config}


class TimeDelayLayer1D(_TimeDelayLayer):
    def __init__(self,
                 filters,
                 kernel_size=5,
                 strides=1,
                 dilation_rate=1,
                 padding='same',
                 activation=None,
                 use_bias=True,
                 **kwargs):
        super(TimeDelayLayer1D, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)

        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.filters = filters
        self.padding = padding

    def build(self, input_shape):
        self.input_dim = int(input_shape[-1])
        self.build_kernel([self.input_dim * self.kernel_size, self.filters])
        self.build_bias(self.filters)
        super(TimeDelayLayer1D, self).build(input_shape)

    def call(self, x, training=None):
        x = time_delay_nn_1d(x, self.get_weight('kernel', training=training),
                             kernel_size=self.kernel_size,
                             strides=self.strides,
                             dilation_rate=self.dilation_rate)
        return super(TimeDelayLayer1D, self).call(x, training=training)

    def get_config(self):
        config = {
            'filters': self.filters,
            'kernel_size': self.kernel_size,
            'strides': self.strides,
            'padding': self.padding,
        }
        base_config = super(TimeDelayLayer1D, self).get_config()
        return {**base_config, **config}


class _MultiTimeDelayLayer(ActivatedKernelBiasBaseLayer):

    def __init__(self,
                 activation=None,
                 use_bias=True,
                 trainable=True,
                 name=None,
                 **kwargs):
        super(_MultiTimeDelayLayer, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            trainable=trainable,
            name=name,
            **kwargs)

    def build_child_layer(self, filters):
        return TimeDelayLayer1D(filters,
                                kernel_size=self.kernel_size,
                                strides=self.strides,
                                dilation_rate=self.dilation_rate,
                                padding=self.padding,
                                use_bias=False,
                                kernel_initializer=self.get_initializer('kernel'),
                                kernel_regularizer=self.get_regularizer('kernel'),
                                kernel_constraint=self.get_constraint('kernel'))

    def get_config(self):
        config = {}
        base_config = super(_MultiTimeDelayLayer, self).get_config()
        return {**base_config, **config}


class DepthGroupwiseTimeDelayLayer1D(_MultiTimeDelayLayer):
    def __init__(self,
                 depth_multiplier=1,
                 kernel_size=5,
                 strides=1,
                 group_size=1,
                 dilation_rate=1,
                 grouped=False,
                 padding='same',
                 activation=None,
                 use_bias=True,
                 input_dim=None,
                 **kwargs):
        super(DepthGroupwiseTimeDelayLayer1D, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)
        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.depth_multiplier = depth_multiplier
        self.group_size = group_size
        self.grouped = grouped
        self.padding = padding
        self.input_dim = input_dim
        if input_dim is not None:
            self.init_layers()

    def init_layers(self):
        self.n_groups = self.input_dim // self.group_size
        assert ((self.input_dim % self.group_size) == 0)
        if self.grouped:
            self.layer = super(DepthGroupwiseTimeDelayLayer1D, self).build_child_layer(
                self.depth_multiplier * self.group_size)
        else:
            self.layers = [super(DepthGroupwiseTimeDelayLayer1D, self).build_child_layer(self.depth_multiplier) for _ in
                           range(self.n_groups)]

    def build(self, input_shape):
        if self.input_dim is None:
            self.input_dim = int(input_shape[-1])
            self.init_layers()

        super(DepthGroupwiseTimeDelayLayer1D, self).build_bias(self.input_dim * self.depth_multiplier)
        super(DepthGroupwiseTimeDelayLayer1D, self).build(input_shape)

    def call(self, x, training=None):
        splits = tf.split(x, self.n_groups, -1)
        if self.grouped:
            outs = [self.layer(xs) for xs in splits]
        else:
            outs = [self.layers[i](xs) for i, xs in enumerate(splits)]
        return super(DepthGroupwiseTimeDelayLayer1D, self).call(tf.concat(outs, axis=-1), training=training)

    def get_config(self):
        config = {
            'depth_multiplier': self.depth_multiplier,
            'kernel_size': self.kernel_size,
            'strides': self.strides,
            'group_size': self.group_size,
            'grouped': self.grouped,
            'padding': self.padding,
            'input_dim': self.input_dim,
        }
        base_config = super(DepthGroupwiseTimeDelayLayer1D, self).get_config()
        return {**base_config, **config}


class TimeDelayLayerFake2D(_MultiTimeDelayLayer):
    def __init__(self,
                 filters,
                 kernel_size=5,
                 strides=1,
                 dilation_rate=1,
                 padding='same',
                 activation=None,
                 use_bias=True,
                 input_dim=None,
                 **kwargs):
        super(TimeDelayLayerFake2D, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)
        self.filters = filters
        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.padding = padding
        self.input_dim = input_dim
        self.layers = None
        if self.input_dim is not None:
            self.init_layers()

    def init_layers(self):
        self.layers = [super(TimeDelayLayerFake2D, self).build_child_layer(self.filters) for _ in
                       range(self.input_dim)]

    def build(self, input_shape):
        if self.input_dim is None:
            self.input_dim = int(input_shape[-2])
            self.init_layers()

        super(TimeDelayLayerFake2D, self).build_bias(self.filters)
        super(TimeDelayLayerFake2D, self).build(input_shape)

    def call(self, x, **kwargs):
        splits = tf.split(x, self.input_dim, -2)
        outs = [tf.expand_dims(self.layers[i](tf.squeeze(xs, -2)), axis=-2) for i, xs in enumerate(splits)]
        return super(TimeDelayLayerFake2D, self).call(tf.concat(outs, axis=-2), **kwargs)

    def get_config(self):
        config = {'filters': self.filters, 'kernel_size': self.kernel_size, 'strides': self.strides,
                  'dilation_rate': self.dilation_rate, 'padding': self.padding, 'input_dim': self.input_dim}
        base_config = super(TimeDelayLayerFake2D, self).get_config()
        return {**base_config, **config}


class DepthGroupwiseTimeDelayLayerFake2D(_MultiTimeDelayLayer):
    def __init__(self,
                 depth_multiplier=1,
                 kernel_size=5,
                 strides=1,
                 dilation_rate=1,
                 grouped=False,
                 group_size=1,
                 padding='same',
                 activation=None,
                 use_bias=True,
                 input_dim=None,
                 **kwargs):

        super(DepthGroupwiseTimeDelayLayerFake2D, self).__init__(
            activation=activations.get(activation),
            use_bias=use_bias,
            **kwargs)
        self.depth_multiplier = depth_multiplier
        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation_rate = dilation_rate
        self.padding = padding
        self.layers = None
        self.input_dim = input_dim
        self.grouped = grouped
        self.group_size = group_size
        if self.input_dim is not None:
            self.init_layers()

    def build_child_layer(self, depth_multiplier):
        return DepthGroupwiseTimeDelayLayer1D(depth_multiplier,
                                              grouped=self.grouped,
                                              group_size=self.group_size,
                                              kernel_size=self.kernel_size,
                                              strides=self.strides,
                                              dilation_rate=self.dilation_rate,
                                              padding=self.padding,
                                              use_bias=False,
                                              kernel_initializer=self.get_initializer('kernel'),
                                              kernel_regularizer=self.get_regularizer('kernel'),
                                              kernel_constraint=self.get_constraint('kernel'))

    def init_layers(self):
        self.layers = [self.build_child_layer(self.depth_multiplier) for _ in
                       range(self.input_dim)]

    def build(self, input_shape):
        if self.layers is None:
            self.input_dim = int(input_shape[-2])
            self.init_layers()
        self.input_channels = int(input_shape[-1])
        super(DepthGroupwiseTimeDelayLayerFake2D, self).build_bias(self.input_channels)
        super(DepthGroupwiseTimeDelayLayerFake2D, self).build(input_shape)

    def call(self, x, **kwargs):
        splits = tf.split(x, self.input_dim, -2)
        outs = [tf.expand_dims(self.layers[i](tf.squeeze(xs, -2)), axis=-2) for i, xs in enumerate(splits)]
        return super(DepthGroupwiseTimeDelayLayerFake2D, self).call(tf.concat(outs, axis=-2), **kwargs)

    def get_config(self):
        config = {'depth_multiplier': self.depth_multiplier,
                  'kernel_size': self.kernel_size,
                  'grouped': self.grouped,
                  'group_size': self.group_size,
                  'strides': self.strides,
                  'dilation_rate': self.dilation_rate,
                  'padding': self.padding,
                  'input_dim': self.input_dim}
        base_config = super(DepthGroupwiseTimeDelayLayerFake2D, self).get_config()
        return {**base_config, **config}


del absolute_import
del division
del print_function



import abc

import numpy as np
import tensorflow as tf
import tensorflow_addons.layers as tfal

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops.t2t_common import shape_list


class PoolNormalization2D(DynastesBaseLayer):
    def __init__(self,
                 pool_size=(2, 2),
                 method=tf.image.ResizeMethod.BILINEAR, antialias=True,
                 **kwargs):
        super(PoolNormalization2D, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.method = method
        self.antialias = antialias

    def call(self, x, training=None):
        orig_dtype = x.dtype
        x = tf.cast(x, tf.float32)
        x_size = shape_list(x)[1:-1]
        x_size = np.where(np.array(list(self.pool_size)) == -1, 1, x_size).tolist()
        x_size = tf.convert_to_tensor(x_size)
        reduce_axes = np.where(np.array(self.pool_size) == -1)[0].tolist()
        if len(reduce_axes) == 2:
            x -= tf.reduce_mean(x, axis=reduce_axes, keepdims=True)
            x *= tf.math.rsqrt(tf.reduce_mean(tf.square(x), axis=reduce_axes, keepdims=True) + 1e-8)
            x = tf.cast(x, orig_dtype)
            return x
        pool_size_t = tf.convert_to_tensor(self.pool_size)
        pool_size_t = tf.maximum(pool_size_t, 1)
        pooled_size = x_size // pool_size_t

        def pool_reduce(x, dtype=tf.float32):
            if len(reduce_axes) > 0:
                x = tf.reduce_mean(x, axis=reduce_axes, keepdims=True)
            x = tf.cast(tf.image.resize(tf.image.resize(x,
                                                        pooled_size,
                                                        method=self.method,
                                                        antialias=self.antialias),
                                        x_size,
                                        method=self.method,
                                        antialias=self.antialias), dtype)
            return x

        x -= pool_reduce(x, tf.float32)
        x *= tf.math.rsqrt(pool_reduce(tf.square(x), tf.float32) + 1e-8)
        x = tf.cast(x, orig_dtype)
        return x

    def get_config(self):
        config = {
            'pool_size': self.pool_size,
            'method': self.method,
            'antialias': self.antialias,
        }
        base_config = super(PoolNormalization2D, self).get_config()
        return {**base_config, **config}


class MultiNormalization(DynastesBaseLayer):
    def __init__(self,
                 layers,
                 **kwargs):
        super(MultiNormalization, self).__init__(**kwargs)
        self.norm_layers = layers

    def build(self, input_shape):
        for layer in self.norm_layers:
            layer.build(input_shape)
        self.add_weight('balance', [input_shape[-1], len(self.norm_layers)], trainable=True,
                        constraint=lambda x: tf.math.softmax(x, axis=-1))
        super(MultiNormalization, self).build(input_shape)

    def call(self, inputs, training=None):
        x = tf.stack([layer(inputs, training=training) for layer in self.norm_layers], axis=-1)
        x *= self.get_weight('balance', training=training)
        x = tf.reduce_sum(x, axis=-1)
        return x


class _AdaptiveNormalization(DynastesBaseLayer, abc.ABC):

    def __init__(self, method=tf.image.ResizeMethod.BILINEAR, antialias=True, mode=None, **kwargs):
        super(_AdaptiveNormalization, self).__init__(**kwargs)
        self.method = method
        self.antialias = antialias
        self.mode = mode

    def build(self, input_shape):
        assert isinstance(input_shape, list)
        if len(input_shape) == 3:
            assert (input_shape[1][-1] + input_shape[2][-1]) // 2 == input_shape[0][-1]
            if self.mode is not None:
                assert self.mode == 'provided_mean_var'
            else:
                self.mode = 'provided_mean_var'
        elif len(input_shape) == 2:
            if (input_shape[1][-1] // 2) != input_shape[0][-1]:
                if self.mode is not None:
                    assert self.mode == 'mapped'
                else:
                    self.mode = 'mapped'
                self.add_weight('map_kernel', shape=[input_shape[1][-1], input_shape[0][-1] * 2])
            else:
                if self.mode is not None:
                    assert self.mode == 'provided_meanvar_fused'
                else:
                    self.mode = 'provided_meanvar_fused'
        else:
            raise ValueError('Incorrect input shapes')
        super(_AdaptiveNormalization, self).build(input_shape)

    def call_mod(self, x, mean_var, training=None, mask=None):
        mean_var_shape = shape_list(mean_var)
        if self.mode == 'provided_mean_var':
            mean_var = tf.concat([mean_var], axis=-1)
        elif self.mode == 'mapped':
            mean_var = tf.matmul(mean_var[0], self.get_weight('map_kernel', training=training))
        elif self.mode != 'provided_meanvar_fused':
            raise ValueError('Something is wrong')
        mean_var_shape = shape_list(mean_var)
        x_shape = shape_list(x)
        if len(mean_var_shape) != len(x_shape):
            mean_var = tf.reshape(mean_var, [-1, ] + ([1] * (len(x.shape) - 2)) + [2, x.shape[-1]])
        else:
            mean_var_shape_npa = np.asarray(mean_var_shape[1:-1])
            x_shape_npa = np.asarray(x_shape[1:-1])
            compatible = np.all(np.logical_or(mean_var_shape_npa == 1, mean_var_shape_npa == x_shape_npa))
            if not compatible:
                size = np.where(mean_var_shape_npa == 1, mean_var_shape_npa, x_shape_npa).tolist()
                if len(mean_var_shape) == 4:
                    mean_var = tf.image.resize(mean_var, size, method=self.method)
                elif len(mean_var_shape) == 3:
                    mean_var = tf.squeeze(tf.image.resize(tf.expand_dims(mean_var, 1), [1] + size, method=self.method,
                                                          antialias=self.antialias), axis=1)
                else:
                    raise ValueError('Only works for 1D or 2D tensors')
            shape = [mean_var_shape[0]] + np.where(mean_var_shape_npa == 1, mean_var_shape_npa,
                                                   x_shape_npa).tolist() + [2, x.shape[-1]]
            mean_var = tf.reshape(mean_var, shape)
        mean, var = tf.unstack(mean_var, axis=-2)

        return (x * var) + mean

    def get_config(self):
        config = {
            'method': self.method,
            'antialias': self.antialias,
            'mode': self.mode,
        }
        base_config = super(_AdaptiveNormalization, self).get_config()
        return {**base_config, **config}


class AdaptiveNormalization(_AdaptiveNormalization):
    def __init__(self,
                 norm_layer,
                 **kwargs):
        super(AdaptiveNormalization, self).__init__(**kwargs)
        self.norm_layer = norm_layer

    def build(self, input_shape):
        assert isinstance(input_shape, list)
        self.norm_layer.build(input_shape[0])
        super(AdaptiveNormalization, self).build(input_shape)

    def call(self, inputs, training=None):
        orig_dtype = inputs[0].dtype
        normalized = self.norm_layer(inputs[0], training=training)
        return tf.cast(self.call_mod(normalized, mean_var=inputs[1:], training=training), orig_dtype)


class AdaptiveGroupNormalization(AdaptiveNormalization):

    def __init__(self, n_groups=2,
                 **kwargs):
        super(AdaptiveGroupNormalization, self).__init__(
            tfal.normalizations.GroupNormalization(groups=n_groups, center=False, scale=False), **kwargs)
        self.n_groups = n_groups

    def get_config(self):
        config = {
            'n_groups': self.n_groups,
        }
        base_config = super(AdaptiveGroupNormalization, self).get_config()
        return {**base_config, **config}


class AdaptiveInstanceNormalization(AdaptiveGroupNormalization):

    def __init__(self,
                 **kwargs):
        kwargs["groups"] = -1
        super(AdaptiveInstanceNormalization, self).__init__(**kwargs)


class AdaptiveLayerNormalization(AdaptiveGroupNormalization):

    def __init__(self,
                 **kwargs):
        kwargs["groups"] = 1
        super(AdaptiveLayerNormalization, self).__init__(**kwargs)


class AdaptiveMultiNormalization(AdaptiveNormalization):

    def __init__(self,
                 layers,
                 **kwargs):
        super(AdaptiveMultiNormalization, self).__init__(MultiNormalization(layers), **kwargs)


class AdaptiveLayerInstanceNormalization(AdaptiveMultiNormalization):

    def __init__(self,
                 **kwargs):
        kwargs["layers"] = [tfal.normalizations.GroupNormalization(groups=1, center=False, scale=False),
                            tfal.normalizations.GroupNormalization(groups=-1, center=False, scale=False)]
        super(AdaptiveLayerInstanceNormalization, self).__init__(**kwargs)



import numpy as np
import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops.signal import window_ops

from dynastes.core.nn import array_ops as d_array_ops
from dynastes.ops.t2t_common import shape_list
"""
    STFT, Mel-spectrogram and inverse methods
    Borrows heavily from
    https://github.com/tensorflow/magenta/tree/master/magenta/models/gansynth/lib
    But supports multi-channel input, ie: [..., time, channels] -> [..., time, mels, channels * 2]
"""


def inverse_stft_window_fn(frame_step,
                           forward_window_fn=window_ops.hann_window,
                           name=None):
    """Generates a window function that can be used in `inverse_stft`.
    Constructs a window that is equal to the forward window with a further
    pointwise amplitude correction.  `inverse_stft_window_fn` is equivalent to
    `forward_window_fn` in the case where it would produce an exact inverse.
    See examples in `inverse_stft` documentation for usage.
    Args:
      frame_step: An integer scalar `Tensor`. The number of samples to step.
      forward_window_fn: window_fn used in the forward transform, `stft`.
      name: An optional name for the operation.
    Returns:
      A callable that takes a window length and a `dtype` keyword argument and
        returns a `[window_length]` `Tensor` of samples in the provided datatype.
        The returned window is suitable for reconstructing original waveform in
        inverse_stft.
    """
    with ops.name_scope(name, 'inverse_stft_window_fn', [forward_window_fn]):
        frame_step = ops.convert_to_tensor(frame_step, name='frame_step')
        frame_step.shape.assert_has_rank(0)

    def inverse_stft_window_fn_inner(frame_length, dtype):
        """Computes a window that can be used in `inverse_stft`.
        Args:
          frame_length: An integer scalar `Tensor`. The window length in samples.
          dtype: Data type of waveform passed to `stft`.
        Returns:
          A window suitable for reconstructing original waveform in `inverse_stft`.
        Raises:
          ValueError: If `frame_length` is not scalar, `forward_window_fn` is not a
          callable that takes a window length and a `dtype` keyword argument and
          returns a `[window_length]` `Tensor` of samples in the provided datatype
          `frame_step` is not scalar, or `frame_step` is not scalar.
        """
        with ops.name_scope(name, 'inverse_stft_window_fn', [forward_window_fn]):
            frame_length = ops.convert_to_tensor(frame_length, name='frame_length')
            frame_length.shape.assert_has_rank(0)

            # Use equation 7 from Griffin + Lim.
            forward_window = forward_window_fn(frame_length, dtype=dtype)
            denom = math_ops.square(forward_window)
            overlaps = -(-frame_length // frame_step)  # Ceiling division.
            denom = array_ops.pad(denom, [(0, overlaps * frame_step - frame_length)])
            denom = array_ops.reshape(denom, [overlaps, frame_step])
            denom = math_ops.reduce_sum(denom, 0, keepdims=True)
            denom = array_ops.tile(denom, [overlaps, 1])
            denom = array_ops.reshape(denom, [overlaps * frame_step])
            denom = tf.maximum(denom, 1e-16)
            return forward_window / denom[:frame_length]

    return inverse_stft_window_fn_inner


def eps(tensor_dtype: tf.DType):
    return tf.convert_to_tensor(np.finfo(tensor_dtype.as_numpy_dtype).eps, dtype=tensor_dtype) * 10000


def _safe_log(x: tf.Tensor):
    return tf.math.log(x + eps(x.dtype))


def waves_to_stfts(waves: tf.Tensor, n_fft=512, hop_length=256, discard_dc=True, pad_l=128, pad_r=128) -> tf.Tensor:
    """Convert from waves to complex stfts.
       Args:
         waves: Tensor of the waveform, shape [..., time, channels].
       Returns:
         stfts: Complex64 tensor of stft, shape [..., time, freq, channels].
       """
    stfts = _waves_to_stfts(waves, n_fft=n_fft, hop_length=hop_length, discard_dc=discard_dc, pad_l=pad_l, pad_r=pad_r)
    stfts = tf.squeeze(stfts, axis=-1)  # [..., channels, time, freq]
    stfts_shape = shape_list(stfts)
    perm = list(range(len(stfts_shape)))
    perm = perm[:-3] + [perm[-2], perm[-1], perm[-3]]
    return tf.transpose(stfts, perm=perm)


def stfts_to_waves(stfts: tf.Tensor, n_fft=512, hop_length=256, discard_dc=True, pad_l=128, pad_r=128) -> tf.Tensor:
    """Convert from complex stfts to waves.
    Args:
      stfts: Complex64 tensor of stft, shape [..., time, freq, channels].
    Returns:
      waves: Tensor of the waveform, shape [..., time, channels].
    """
    stfts_shape = shape_list(stfts)
    perm = list(range(len(stfts_shape)))
    perm = perm[:-3] + [perm[-1], perm[-3], perm[-2]]
    stfts = tf.transpose(stfts, perm=perm)  # [..., channels, time, freq]
    stfts = tf.expand_dims(stfts, axis=-1)

    waves = _stfts_to_waves(stfts, n_fft=n_fft, hop_length=hop_length, discard_dc=discard_dc, pad_l=pad_l,
                            pad_r=pad_r)  # [..., channels, time, freq]
    return waves


def stfts_to_melspecgrams(stfts: tf.Tensor, l2mel, ifreq=True) -> tf.Tensor:
    """Converts stfts to specgrams.
    Args:
      stfts: Complex64 tensor of stft, shape [..., time, freq, channels].
    Returns:
      melspecgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., time, freq, 2*channels], mel scaling of frequencies.
    """
    # inp: [..., time, freq, channels]
    stfts_shape = shape_list(stfts)
    perm = list(range(len(stfts_shape)))
    perm = perm[:-3] + [perm[-1], perm[-3], perm[-2]]
    stfts = tf.transpose(stfts, perm=perm)
    stfts = tf.expand_dims(stfts, axis=-1)  # [..., channels, time, freq, 1]
    melspecgrams = _stfts_to_melspecgrams(stfts, l2mel=l2mel, ifreq=ifreq)  # [..., channels, time, freq, 2]

    melspecgrams_shape = shape_list(melspecgrams)
    perm = list(range(len(melspecgrams_shape)))
    perm = perm[:-4] + [perm[-3], perm[-2], perm[-4], perm[-1]]
    melspecgrams = tf.transpose(melspecgrams, perm=perm)  # [..., time, freq, channels, 2]
    melspecgrams_shape = shape_list(melspecgrams)
    melspecgrams = tf.reshape(melspecgrams, melspecgrams_shape[:-2] + [
        melspecgrams_shape[-2] * melspecgrams_shape[-1]])  # [..., time, freq, channels * 2]
    return melspecgrams


def melspecgrams_to_stfts(melspecgrams: tf.Tensor, mel2l, ifreq=True) -> tf.Tensor:
    """Converts melspecgrams to stfts.
        Args:
          melspecgrams: Tensor of log magnitudes and instantaneous frequencies,
            shape [..., time, freq, 2*channels], mel scaling of frequencies.
          mel2l: Mel to linear matrix, ie transposed linear to mel matrix
            @see
            tf.signal.linear_to_mel_weight_matrix
        Returns:
          specgrams: Tensor of log magnitudes and instantaneous frequencies,
            shape [..., time, freq, channels].
        """
    melspecgrams_shape = shape_list(melspecgrams)  # [..., time, freq, channels*2]
    melspecgrams = tf.reshape(melspecgrams,
                              melspecgrams_shape[:-1] + [melspecgrams_shape[-1] // 2,
                                                         2])  # [..., time, freq, channels, 2]
    perm = list(range(len(melspecgrams_shape) + 1))
    perm = perm[:-4] + [perm[-2], perm[-4], perm[-3], perm[-1]]
    melspecgrams = tf.transpose(melspecgrams, perm=perm)  # [..., channels, time, freq, 2]
    stfts = _melspecgrams_to_stfts(melspecgrams, mel2l=mel2l, ifreq=True)  # [..., channels, time, freq, 1]
    stfts = tf.squeeze(stfts, axis=-1)  # [..., channels, time, freq]
    stfts_shape = shape_list(stfts)
    perm = list(range(len(stfts_shape)))
    perm = perm[:-3] + [perm[-2], perm[-1], perm[-3]]
    stfts = tf.transpose(stfts, perm=perm)  # [..., time, freq, channels]
    return stfts


def _waves_to_stfts(waves: tf.Tensor, n_fft=512, hop_length=256, discard_dc=True, pad_l=128, pad_r=128) -> tf.Tensor:
    """Convert from waves to complex stfts.
    Args:
      waves: Tensor of the waveform, shape [..., time, channels].
    Returns:
      stfts: Complex64 tensor of stft, shape [..., channels, time, freq, 1].
    """
    waves_shape = shape_list(waves)
    waves = tf.linalg.matrix_transpose(waves)  # [..., channels, time]
    waves_padded = tf.pad(waves,
                          np.reshape(np.asarray([0, 0] * (len(waves_shape) - 1)), (-1, 2)).tolist() + [[pad_l, pad_r]])
    stfts = tf.signal.stft(
        waves_padded,
        window_fn=tf.signal.hann_window,
        frame_length=n_fft,
        frame_step=hop_length,
        fft_length=n_fft,
        pad_end=False)
    if discard_dc:
        stfts, dc = tf.split(stfts, num_or_size_splits=[n_fft // 2, 1], axis=-1)
    return tf.expand_dims(stfts, axis=-1)


def _stfts_to_waves(stfts, n_fft=512, hop_length=256, discard_dc=True, pad_l=128, pad_r=128):
    """Convert from complex stfts to waves.
    Args:
      stfts: Complex64 tensor of stft, shape [..., channels, time, freq, 1].
    Returns:
      waves: Tensor of the waveform, shape [..., time, channels].
    """
    stfts = tf.squeeze(stfts, axis=-1)
    stfts_shape = shape_list(stfts)
    dc = 1 if discard_dc else 0
    nyq = 1 - dc
    stfts = tf.pad(stfts, np.reshape(np.asarray([0, 0] * (len(stfts_shape) - 1)), (-1, 2)).tolist() + [[dc, nyq]])
    waves_resyn = tf.signal.inverse_stft(
        stfts=stfts,
        frame_length=n_fft,
        frame_step=hop_length,
        fft_length=n_fft,
        window_fn=inverse_stft_window_fn(frame_step=hop_length))
    waves_resyn = tf.linalg.matrix_transpose(waves_resyn)
    crops = np.reshape(np.asarray([0, 0] * (len(stfts_shape) - 3)), (-1, 2)).tolist() + [[pad_l, pad_r], [0, 0]]
    return d_array_ops.crop(waves_resyn, crops)


def _stfts_to_specgrams(stfts, ifreq=True):
    """Converts stfts to specgrams.
    Args:
      stfts: Complex64 tensor of stft, shape [..., channels, time, freq, 1].
    Returns:
      specgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2].
    """
    stfts = tf.squeeze(stfts, axis=-1)
    logmag = _safe_log(tf.abs(stfts))

    phase_angle = tf.math.angle(stfts)
    if ifreq:
        p = instantaneous_frequency(phase_angle)
    else:
        p = phase_angle / np.pi

    return tf.stack([logmag, p], axis=-1)


def _specgrams_to_melspecgrams(specgrams, l2mel):
    """Converts specgrams to melspecgrams.
    Args:
      specgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2].
      l2mel: Linear to mel matrix
    Returns:
      melspecgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2], mel scaling of frequencies.
    """

    logmag, p = tf.unstack(specgrams, axis=-1)

    mag2 = tf.exp(2.0 * logmag)
    phase_angle = tf.cumsum(p * np.pi, axis=-2)

    logmelmag2 = _safe_log(tf.tensordot(mag2, l2mel, 1))
    mel_phase_angle = tf.tensordot(phase_angle, l2mel, 1)
    mel_p = instantaneous_frequency(mel_phase_angle)

    return tf.stack([logmelmag2, mel_p], axis=-1)


def _stfts_to_melspecgrams(stfts: tf.Tensor, l2mel, ifreq=True) -> tf.Tensor:
    """Converts stfts to specgrams.
    Args:
      stfts: Complex64 tensor of stft, shape [..., channels, time, freq, 1].
    Returns:
      melspecgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2], mel scaling of frequencies.
    """
    logmelmag2 = _safe_log(tf.matmul(tf.abs(stfts), l2mel, transpose_a=True))
    logmelmag2 = tf.linalg.matrix_transpose(logmelmag2)
    phase_angle = tf.math.angle(stfts)
    mel_phase_angle = tf.matmul(phase_angle, l2mel, transpose_a=True)
    mel_phase_angle = tf.linalg.matrix_transpose(mel_phase_angle)
    if ifreq:
        mel_p = instantaneous_frequency(mel_phase_angle)
    else:
        mel_p = mel_phase_angle / np.pi
    return tf.concat([logmelmag2, mel_p], axis=-1)


def _specgrams_to_stfts(specgrams: tf.Tensor, ifreq=True) -> tf.Tensor:
    """Converts specgrams to stfts.
    Args:
      specgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2].
    Returns:
      stfts: Complex64 tensor of stft, shape [..., channels, time, freq, 1].
    """
    logmag, p = tf.unstack(specgrams, axis=-1)

    mag = tf.exp(logmag)

    if ifreq:
        phase_angle = tf.cumsum(p * np.pi, axis=-2)
    else:
        phase_angle = p * np.pi

    return tf.expand_dims(polar2rect(mag, phase_angle), axis=-1)


def _melspecgrams_to_specgrams(melspecgrams: tf.Tensor, mel2l) -> tf.Tensor:
    """Converts melspecgrams to specgrams.
    Args:
      melspecgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2], mel scaling of frequencies.
      mel2l: Mel to linear matrix
    Returns:
      specgrams: Tensor of log magnitudes and instantaneous frequencies,
        shape [..., channels, time, freq, 2].
    """

    logmelmag2, mel_p = tf.unstack(melspecgrams, axis=-1)

    mag2 = tf.tensordot(tf.exp(logmelmag2), mel2l, 1)
    logmag = 0.5 * _safe_log(mag2)
    mel_phase_angle = tf.cumsum(mel_p * np.pi, axis=-2)
    phase_angle = tf.tensordot(mel_phase_angle, mel2l, 1)
    p = instantaneous_frequency(phase_angle)

    return tf.stack([logmag, p], axis=-1)


def _melspecgrams_to_stfts(melspecgrams, mel2l, ifreq=True):
    """Converts melspecgrams to specgrams.
        Args:
          melspecgrams: Tensor of log magnitudes and instantaneous frequencies,
            shape [..., channels, time, freq, 2], mel scaling of frequencies.
          mel2l: Mel to linear matrix
        Returns:
          stfts: Tensor of log magnitudes and instantaneous frequencies,
            shape [..., channels, time, freq, 1].
        """
    logmelmag, mel_p = tf.unstack(melspecgrams, axis=-1)

    mag = tf.matmul(tf.exp(logmelmag), mel2l)
    if ifreq:
        mel_phase_angle = tf.cumsum(mel_p * np.pi, axis=-2)
    else:
        mel_phase_angle = mel_p * np.pi

    phase_angle = tf.matmul(mel_phase_angle, mel2l)

    return tf.expand_dims(polar2rect(mag, phase_angle), axis=-1)


def _diff(x, axis=-1):
    """Take the finite difference of a tensor along an axis.
    Args:
      x: Input tensor of any dimension.
      axis: Axis on which to take the finite difference.
    Returns:
      d: Tensor with size less than x by 1 along the difference dimension.
    Raises:
      ValueError: Axis out of range for tensor.
    """
    shape = x.get_shape()
    if axis >= len(shape):
        raise ValueError('Invalid axis index: %d for tensor with only %d axes.' %
                         (axis, len(shape)))

    begin_back = [0 for _ in range(len(shape))]
    begin_front = [0 for _ in range(len(shape))]
    begin_front[axis] = 1

    size = shape.as_list()
    size[axis] -= 1
    slice_front = tf.slice(x, begin_front, size)
    slice_back = tf.slice(x, begin_back, size)
    d = slice_front - slice_back
    return d


def _unwrap(p, axis=-1):
    """Unwrap a cyclical phase tensor.
    Args:
      p: Phase tensor.
      discont: Float, size of the cyclic discontinuity.
      axis: Axis of which to unwrap.
    Returns:
      unwrapped: Unwrapped tensor of same size as input.
    """
    dd = _diff(p, axis=axis)
    ddmod = tf.math.mod(dd + np.pi, 2.0 * np.pi) - np.pi
    idx = tf.logical_and(tf.equal(ddmod, -np.pi), tf.greater(dd, 0))
    ddmod = tf.where(idx, tf.ones_like(ddmod) * np.pi, ddmod)
    ph_correct = ddmod - dd
    ph_cumsum = tf.cumsum(ph_correct, axis=axis)

    shape = p.get_shape().as_list()
    shape[axis] = 1
    ph_cumsum = tf.concat([tf.zeros(shape, dtype=p.dtype), ph_cumsum], axis=axis)
    unwrapped = p + ph_cumsum
    return unwrapped


def instantaneous_frequency(phase_angle, time_axis=-2, use_unwrap=True):
    """Transform a fft tensor from phase angle to instantaneous frequency.
    Take the finite difference of the phase. Pad with initial phase to keep the
    tensor the same size.
    Args:
      phase_angle: Tensor of angles in radians. [..., Time, Freqs]
      time_axis: Axis over which to unwrap and take finite difference.
      use_unwrap: True preserves original GANSynth behavior, whereas False will
          guard against loss of precision.
    Returns:
      dphase: Instantaneous frequency (derivative of phase). Same size as input.
    """
    if use_unwrap:
        # Can lead to loss of precision.
        phase_unwrapped = _unwrap(phase_angle, axis=time_axis)
        dphase = _diff(phase_unwrapped, axis=time_axis)
    else:
        # Keep dphase bounded. N.B. runs faster than a single mod-2pi expression.
        dphase = _diff(phase_angle, axis=time_axis)
        dphase = tf.where(dphase > np.pi, dphase - 2 * np.pi, dphase)
        dphase = tf.where(dphase < -np.pi, dphase + 2 * np.pi, dphase)

    # Add an initial phase to dphase.
    size = phase_angle.get_shape().as_list()
    size[time_axis] = 1
    begin = [0 for _ in size]
    phase_slice = tf.slice(phase_angle, begin, size)
    dphase = tf.concat([phase_slice, dphase], axis=time_axis) / np.pi
    return dphase


def polar2rect(mag, phase_angle):
    """Convert polar-form complex number to its rectangular form."""
    mag = tf.complex(mag, tf.convert_to_tensor(0.0, dtype=mag.dtype))
    phase = tf.complex(tf.cos(phase_angle), tf.sin(phase_angle))
    return mag * phase



import os
import codecs
import re
from setuptools import setup, find_packages

current_path = os.path.abspath(os.path.dirname(__file__))


def read_file(*parts):
    with codecs.open(os.path.join(current_path, *parts), 'r', 'utf8') as reader:
        return reader.read()

def get_requirements(*parts):
    with codecs.open(os.path.join(current_path, *parts), 'r', 'utf8') as reader:
        return list(map(lambda x: x.strip(), reader.readlines()))

def find_version(*file_paths):
    version_file = read_file(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


setup(
    name="dynastes",
    version=find_version('dynastes', '__init__.py'),
    author="Göran Sandström (Veqtor)",
    author_email="v3qt0r@gmail.com",
    description="A collection of layers and utils for TensorFlow (Keras) 2.+",
    long_description=read_file('README.md'),
    long_description_content_type="text/markdown",
    url="https://github.com/veqtor/dynastes",
    install_requires=get_requirements('requirements.txt'),
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import tensorflow as tf


def to_float(x):
    """Cast x to float; created because to_float is deprecated."""
    return tf.cast(x, tf.float32)


def to_int32(x):
    """Cast x to float; created because to_float is deprecated."""
    return tf.cast(x, tf.int32)


class TruncatingDispatcher(object):
    """Helper for implementing a mixture of experts.
    A TruncatingDispatcher is useful when you need to deal with
    fixed-sized Tensors.  As opposed to a SparseDispatcher, which
    produces batches of different sizes for the different experts, the
    TruncatingDispatcher always produces batches of the same given size,
    and the results are returned stacked in one big tensor.
    In the case where an expert is over-capacity, the last items that
    should have gone to that expert are dropped.
    Confusingly, the inputs to a TruncatingDispatcher have both a
    "batch" and a "length" dimension.  Not only does each expert receive
    the same total number of examples, it also receives the same number
    of examples for each element of "batch".  This behavior is necessary
    for applications such as grouped attention, where we have a batch of
    sequences, and we want each sequence to be divided evenly among
    experts.  For simpler applications like mixture-of-experts, you can
    reshape the input so that the "batch" dimension is 1, and only the
    "length" dimension is used.
    """

    def __init__(self, requests, expert_capacity):
        """Create a TruncatingDispatcher.
        Args:
          requests: a boolean `Tensor` of shape `[batch, length, num_experts]`.
            Alternatively, a float or int Tensor containing zeros and ones.
          expert_capacity: a Scalar - maximum number of examples per expert per
            batch element.
        Returns:
          a TruncatingDispatcher
        """
        self._requests = to_float(requests)
        self._expert_capacity = expert_capacity
        expert_capacity_f = to_float(expert_capacity)
        self._batch, self._length, self._num_experts = tf.unstack(
            tf.shape(self._requests), num=3)

        # [batch, length, num_experts]
        position_in_expert = tf.cumsum(self._requests, axis=1, exclusive=True)
        # [batch, length, num_experts]
        self._gates = self._requests * to_float(
            tf.less(position_in_expert, expert_capacity_f))
        batch_index = tf.reshape(
            to_float(tf.range(self._batch)), [self._batch, 1, 1])
        length_index = tf.reshape(
            to_float(tf.range(self._length)), [1, self._length, 1])
        expert_index = tf.reshape(
            to_float(tf.range(self._num_experts)), [1, 1, self._num_experts])
        # position in a Tensor with shape [batch * num_experts * expert_capacity]
        flat_position = (
                position_in_expert +
                batch_index * (to_float(self._num_experts) * expert_capacity_f) +
                expert_index * expert_capacity_f)
        # Tensor of shape [batch * num_experts * expert_capacity].
        # each element is an integer in [0, length)
        self._indices = tf.math.unsorted_segment_sum(
            data=tf.reshape((length_index + 1.0) * self._gates, [-1]),
            segment_ids=to_int32(tf.reshape(flat_position, [-1])),
            num_segments=self._batch * self._num_experts * expert_capacity)
        self._indices = tf.reshape(
            self._indices,
            [self._batch, self._num_experts, expert_capacity])
        # Tensors of shape [batch, num_experts, expert_capacity].
        # each element is 0.0 or 1.0
        self._nonpadding = tf.minimum(self._indices, 1.0)
        # each element is an integer in [0, length)
        self._indices = tf.nn.relu(self._indices - 1.0)
        # self._flat_indices is [batch, num_experts, expert_capacity], with values
        # in [0, batch * length)
        self._flat_indices = to_int32(
            self._indices +
            (tf.reshape(to_float(tf.range(self._batch)), [-1, 1, 1])
             * to_float(self._length)))
        self._indices = to_int32(self._indices)

    def dispatch(self, inp):
        """Send the inputs to the experts.
        Args:
          inp: a `Tensor` of shape "[batch, length, depth]`
        Returns:
          a tensor with shape [batch, num_experts, expert_capacity, depth]
        """
        inp = tf.reshape(inp, [self._batch * self._length, -1])
        # [batch, num_experts, expert_capacity, depth]
        ret = tf.gather(inp, self._flat_indices)
        return ret

    def combine(self, x):
        """Return the output from the experts.
        When one example goes to multiple experts, the outputs are summed.
        Args:
          x: a Tensor with shape [batch, num_experts, expert_capacity, depth]
        Returns:
          a `Tensor` with shape `[batch, length, depth]
        """
        depth = tf.shape(x)[-1]
        x *= tf.expand_dims(self._nonpadding, -1)
        ret = tf.math.unsorted_segment_sum(
            x, self._flat_indices, num_segments=self._batch * self._length)
        ret = tf.reshape(ret, [self._batch, self._length, depth])
        return ret

    def nonpadding(self):
        """Which elements of a dispatched Tensor are not padding.
        Returns:
          a Zero/One float tensor with shape [batch, num_experts, expert_capacity].
        """
        return self._nonpadding

    def gates(self):
        """A Tensor indicating which examples go to which experts.
        Returns:
          A float32 Tensor with shape [batch, length, num_experts], where each value
          is 0.0 or 1.0.
        """
        return self._gates

    def length_coordinate(self):
        """Length coordinate of dispatched tensor.
        Returns:
          a tensor with shape [batch, num_experts, expert_capacity] containing
           integers in the range [0, length)
        """
        return self._indices



from abc import ABCMeta
from functools import partial
from typing import List

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

tfpb = tfp.bijectors


def _get_roll_op(axis=0, steps=1, constant=0):
    def roll_op(x):
        return tf.stack([tf.roll(_x, -((i * steps) + constant), axis) for i, _x in enumerate(tf.unstack(x, axis=axis))],
                        axis=axis)

    return roll_op


class RollIncremental(tfpb.Bijector, metaclass=ABCMeta):

    def __init__(
            self,
            axis,
            steps=0,
            constant=0,
            **kwargs):
        kwargs['forward_min_event_ndims'] = axis + 1
        kwargs['inverse_min_event_ndims'] = axis + 1
        super(RollIncremental, self).__init__(**kwargs)
        setattr(self, '_forward', _get_roll_op(axis=axis, steps=steps, constant=constant))
        setattr(self, '_inverse', _get_roll_op(axis=axis, steps=-steps, constant=-constant))
        setattr(self, 'forward_event_shape', lambda x: x)
        setattr(self, 'inverse_event_shape', lambda x: x)


class EventShapeAwareChain(tfpb.Chain):

    def __init__(self,
                 event_shape_in,
                 partial_bijectors: List[partial],
                 **kwargs):

        bijectors = []
        for partial_bijector in partial_bijectors:
            if partial_bijector.func == tfpb.Reshape:
                if partial_bijector.keywords['event_shape_out'][-1] == -1:
                    if np.cumprod(np.array(partial_bijector.keywords['event_shape_out'][:-1]))[-1] == \
                            np.cumprod(np.array(event_shape_in[:-1]))[-1]:
                        partial_bijector.keywords['event_shape_out'][-1] = event_shape_in[-1]
                bijector = partial_bijector(event_shape_in=event_shape_in)
            else:
                bijector = partial_bijector()
            event_shape_in = bijector.forward_event_shape(event_shape_in)
            bijectors.append(bijector)
        kwargs['bijectors'] = list(reversed(bijectors))
        super(EventShapeAwareChain, self).__init__(**kwargs)



from functools import partial

import tensorflow_probability as tfp

tfpb = tfp.bijectors
from dynastes.probability import bijectors


def Reshape(event_shape_out) -> partial:
    return partial(tfpb.Reshape, event_shape_out=event_shape_out)


def RollIncremental(axis,
                    steps=0,
                    constant=0) -> partial:
    return partial(bijectors.RollIncremental, axis=axis, steps=steps, constant=constant)


def Transpose(perm) -> partial:
    return partial(tfpb.Transpose, perm=perm)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.keras import layers as tfkl

from dynastes import layers as vqkl


def get_1d_layer(type,
                 filters,
                 depth_multiplier,
                 kernel_size,
                 strides=1,
                 dilation_rate=1,
                 grouped=False,
                 group_size=1,
                 padding='same',
                 activation=None,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):
    if type.lower() == 'TimeDelayLayer1D'.lower():
        return vqkl.TimeDelayLayer1D(filters=filters,
                                     kernel_size=kernel_size,
                                     strides=strides,
                                     dilation_rate=dilation_rate,
                                     padding=padding,
                                     activation=activation,
                                     use_bias=use_bias,
                                     kernel_initializer=kernel_initializer,
                                     bias_initializer=bias_initializer,
                                     kernel_regularizer=kernel_regularizer,
                                     bias_regularizer=bias_regularizer,
                                     activity_regularizer=activity_regularizer,
                                     kernel_constraint=kernel_constraint,
                                     bias_constraint=bias_constraint, **kwargs)
    elif type.lower() == 'DepthGroupwiseTimeDelayLayer1D'.lower():
        return vqkl.DepthGroupwiseTimeDelayLayer1D(depth_multiplier=depth_multiplier,
                                                   kernel_size=kernel_size,
                                                   strides=strides,
                                                   dilation_rate=dilation_rate,
                                                   padding=padding,
                                                   activation=activation,
                                                   use_bias=use_bias,
                                                   grouped=grouped,
                                                   group_size=group_size,
                                                   kernel_initializer=kernel_initializer,
                                                   bias_initializer=bias_initializer,
                                                   kernel_regularizer=kernel_regularizer,
                                                   bias_regularizer=bias_regularizer,
                                                   activity_regularizer=activity_regularizer,
                                                   kernel_constraint=kernel_constraint,
                                                   bias_constraint=bias_constraint, **kwargs)
    elif type.lower() in ['Convolution1D'.lower(), 'Conv1D'.lower()]:
        return tfkl.Convolution1D(filters=filters,
                                  kernel_size=kernel_size,
                                  strides=strides,
                                  dilation_rate=dilation_rate,
                                  padding=padding,
                                  activation=activation,
                                  use_bias=use_bias,
                                  grouped=grouped,
                                  group_size=group_size,
                                  kernel_initializer=kernel_initializer,
                                  bias_initializer=bias_initializer,
                                  kernel_regularizer=kernel_regularizer,
                                  bias_regularizer=bias_regularizer,
                                  activity_regularizer=activity_regularizer,
                                  kernel_constraint=kernel_constraint,
                                  bias_constraint=bias_constraint, **kwargs)
    elif type.lower() in ['SeparableConv1D'.lower(), 'SeparableConvolution1D'.lower()]:
        return tfkl.SeparableConvolution1D(filters=filters,
                                           kernel_size=kernel_size,
                                           strides=strides,
                                           dilation_rate=dilation_rate,
                                           padding=padding,
                                           activation=activation,
                                           use_bias=use_bias,
                                           grouped=grouped,
                                           group_size=group_size,
                                           kernel_initializer=kernel_initializer,
                                           bias_initializer=bias_initializer,
                                           kernel_regularizer=kernel_regularizer,
                                           bias_regularizer=bias_regularizer,
                                           activity_regularizer=activity_regularizer,
                                           kernel_constraint=kernel_constraint,
                                           bias_constraint=bias_constraint, **kwargs)


def get_1D_attention_layer(type,
                           strides,
                           dilation_rate,
                           num_heads,
                           padding,
                           multiquery_attention,
                           self_attention=True,
                           preshaped_q=True,
                           relative=False,
                           local=False,
                           sparse=False,
                           masked=False,
                           dropout_rate=0.,
                           max_relative_position=None,
                           lsh_bucket_length=4,
                           block_length=128,
                           filter_width=100,
                           mask_right=False,
                           add_relative_to_values=False):
    if type.lower() == 'LocalizedAttentionLayer1D'.lower():
        return vqkl.LocalizedAttentionLayer1D(strides=strides,
                                              dilation_rate=dilation_rate,
                                              num_heads=num_heads,
                                              padding=padding,
                                              preshaped_q=preshaped_q)
    elif type.lower() == 'Attention1D'.lower():
        return vqkl.Attention1D(num_heads=num_heads,
                                multiquery_attention=multiquery_attention,
                                self_attention=self_attention,
                                relative=relative,
                                masked=masked,
                                sparse=sparse,
                                local=local,
                                dropout_rate=dropout_rate,
                                max_relative_position=max_relative_position,
                                lsh_bucket_length=lsh_bucket_length,
                                block_length=block_length,
                                filter_width=filter_width,
                                mask_right=mask_right,
                                add_relative_to_values=add_relative_to_values)
